import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

// an arbitrarily large number is represented by a list of digits and a sign
public class BigNum {
    private static final int BASE = 10; // could be larger, just beware of overflow on carries.  10 right now for ease of understanding
    private static final String DIGITS = "0123456789"; // for the string constructor, what digits are allowed
    private List<Integer> digits; // units digit is at the start of the list, no leading (most-significant) zeros as an invariant
    								// therefore 0 is an empty list with an arbitrary sign
    private boolean isNegated; // false means positive

    public BigNum() {
        digits = new LinkedList<Integer>(); // could have made it a new ArrayList<Integer>() too
    }

	// converts a string into a number, where the number consists of digits and an optional - prefix
	// commas are ignored.  a number in any other format is garbage-in garbage-out
    public BigNum(String s) {
		this();
		if (s.length()==0) return;
        for(int i = s.length()-1; i>=0; i--) {
            char c = s.charAt(i);
            int idx = DIGITS.indexOf(c);
            if (idx != -1) digits.add(idx);
        }
        if (s.charAt(0) == '-') isNegated = true;
	stripLeadingZeros();
    }

	// positive +1, negative -1, zero 0
	public int getSign() {
		if (digits.isEmpty()) return 0;
		else if (isNegated()) return -1;
		else return 1;
	}

// true if the numbers are equivalent.  tricky case is +0 is equivalent to -0
	public boolean equals(Object o) {
		throw new UnsupportedOperationException("you need to write this code");
	}

	// FINISHME make it separate every three digits with a ',' e.g. 4,321 and 654,321 and 87,654,321
	public String toString() {
		String s = "";
		if (digits.isEmpty()) return "0"; // special case for 0
		for(Integer i:digits) { // example of a closure
		    s = i.intValue() + s;
		}
		if (isNegated()) s = "-"+s;
		return s;
	}

	// negates (modifies) this BigNum
	public BigNum negateMe() {
		isNegated = !isNegated;
		return this;
	}

	// multiply returns a new BigNum storing the product of this * other
	// you need to have the different cases of positive and negative here, much like add.
	public BigNum multiply(BigNum other) {
	    throw new UnsupportedOperationException("you need to write this code");
	}

	// adds two arbitrarily signed numbers by breaking the signs into cases so the arithmetic
	// can all be done on positives
    public BigNum add(BigNum other) {
		if (isNegated() == other.isNegated()) { // signs the same
			BigNum sum = addPositives(other);
			if (isNegated()) return sum.negateMe();
			else return sum;
		} else { // signs different
		    BigNum diff;
			if (biggerMagnitudeThan(other)) {
			    diff = subtractSmallerFromBigger(other);
			    diff.stripLeadingZeros();
			    if (isNegated()) return diff.negateMe();
			    else return diff;
			} else {
			    diff = other.subtractSmallerFromBigger(this);
			    diff.stripLeadingZeros();
			    if (other.isNegated()) return diff.negateMe();
			    else return diff;
			}
		}
	}

	// reduces a subtraction to an addition so the cases don't need to be enumerated
    public BigNum subtract(BigNum other) {
		other = other.add(new BigNum()); // hack, make a new copy
		other.negateMe(); // earlier version broke on a.subtract(a)
		return add(other);
	}

	// signs are ignored, the sum is returned as a new BigNum
    private BigNum addPositives(BigNum other) {
        BigNum sum=new BigNum();
        int carry=0;
        Iterator<Integer> it1 = digits.iterator();
        Iterator<Integer> it2 = other.digits.iterator();
	    while(it1.hasNext() || it2.hasNext()) {
            int a,b;
	        if (it1.hasNext()) a = it1.next();
	        else a = 0;
	        if (it2.hasNext()) b = it2.next();
	        else b = 0;
	        int c = a+b+carry;
	        carry = c/BASE;
	        sum.digits.add(c%BASE);
	    }
		if (carry!=0) sum.digits.add(carry);
	    return sum;
    }

	// subtracts other from this.
	// the sign is ignored.  precondition is that this is bigger than other.
	// the returned difference is a new BigNum that is nonnegative
	// don't worry about leading zeros -- they'll be stripped later
    private BigNum subtractSmallerFromBigger(BigNum other) {
		throw new UnsupportedOperationException("you need to write this code");
	}

	// multiplies two numbers as if they were positive, returning the product as a new BigNum
    private BigNum multiplyPositives(BigNum other) {
        BigNum prod = new BigNum();
        if (this.digits.isEmpty()) return prod;
        int k = 0;
		for(Integer i:other.digits) {
			prod=prod.add(this.multShift(i,k));
			k++;
		}
		return prod;
	}

	// helper method for multiplication, in effect doing one row of the multiply
	private BigNum multShift(int digit, int shift) {
		BigNum ans = new BigNum();
		for(int i = 0; i<shift; i++)
		    ans.digits.add(0);
		int carry=0;
		for(Integer i:digits) {
			int p = i*digit+carry;
			carry = p/BASE;
			ans.digits.add(p%BASE);
		}
		if (carry!=0) ans.digits.add(carry);
		return ans;
	}

	// this method assumes no leading zeros
	// returns true if this is bigger magnitude than other
	private boolean biggerMagnitudeThan(BigNum other) {
		if (digits.size()>other.digits.size()) return true;
		else if (digits.size()<other.digits.size()) return false;
		else { // equal numbers of digits
			ListIterator<Integer> li1 = digits.listIterator(digits.size());
			ListIterator<Integer> li2 = other.digits.listIterator(digits.size());
			while(li1.hasPrevious()) {
				int k = li1.previous()-li2.previous();
				if (k>0) return true;
				if (k<0) return false;
			}
		}
		return false; // equal
	}

	// strips the leading zeros by walking backward from the end of the array
	private void stripLeadingZeros() {
        ListIterator<Integer> li = digits.listIterator(digits.size()); // start at the most-significant digit
        while(li.hasPrevious()) {
			if (li.previous() == 0) li.remove();
			else return;
		}
	}

	private boolean isNegated() {
		return isNegated;
	}
}
